/* word.c   by Michael Thorpe   2005-08-12 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lisp.h"

#define HASH_SIZE 127

struct hashentry {
	struct hashentry *prev,*next;
	obj *word;
};

static struct hashentry *hashtable[HASH_SIZE];

static int hash(const char *data,int len) {
	unsigned int h=0,m=1;

	while(len--) {
		h=(h+m*data++[0])%HASH_SIZE;
		m=(m*3)%HASH_SIZE;
	}
	return(h);
}

static void drop_word(obj *o) {
	int h;
	struct hashentry *entry;

	h=hash(o->value.s.data,o->value.s.len);
	for(entry=hashtable[h];entry;entry=entry->next)
		if(entry->word->value.s.len==o->value.s.len && !memcmp(o->value.s.data,entry->word->value.s.data,o->value.s.len))
			break;
	if(!entry) {
		fprintf(stderr,"Just tried to drop a word I don't know about from %d: %.*s\n",h,(int)o->value.s.len,o->value.s.data);
		exit(-1);
	}
	if(entry->prev)
		entry->prev->next=entry->next;
	else
		hashtable[h]=entry->next;
	if(entry->next)
		entry->next->prev=entry->prev;
	free(entry);
	free(o->value.s.data);
}

static obj *word_evaluate(scope *scope,obj *in) {
	return(lookup(scope,in));
}

static obj *word_tostring(obj *in) {
	obj *out;

	out=real_new_string(in->value.s.data,in->value.s.len);
	decref(in);
	return(out);
}

obj *newword(const char *data,unsigned long len) {
	obj *o;
	char *s;
	int h;
	struct hashentry *entry;

	h=hash(data,len);
	for(entry=hashtable[h];entry;entry=entry->next)
		if(entry->word->value.s.len==len && !memcmp(data,entry->word->value.s.data,len))
			break;
	if(entry)
		return(incref(entry->word));
	s=(char *)malloc(len);
	if(!s)
		return(throwoom());
	entry=(struct hashentry *)malloc(sizeof(struct hashentry));
	if(!entry) {
		free(s);
		return(throwoom());
	}
	o=newobj(&word_objtype);
	if(!o) {
		free(entry);
		free(s);
		return(0);
	}
	entry->word=o;
	o->value.s.data=s;
	o->value.s.len=len;
	if(data)
		memcpy(o->value.s.data,data,len);
	else
		memset(o->value.s.data,0,len);
	entry->prev=0;
	entry->next=hashtable[h];
	if(hashtable[h])
		hashtable[h]->prev=entry;
	hashtable[h]=entry;
	return(o);
}

obj *unique_word() {
	int h,len;
	struct hashentry *entry;
	char buf[1+4+1+10+1];

	while(1) {
		len=snprintf(buf,sizeof(buf),"#word-%u",rand());
		if(-1==len || sizeof(buf)<=len) {
			fprintf(stderr,"unique_word ust got %d from snprintf\n",len);
			exit(-1);
		}
		h=hash(buf,len);
		for(entry=hashtable[h];entry;entry=entry->next)
			if(entry->word->value.s.len != len || memcmp(buf,entry->word->value.s.data,len)) {
				return(newword(buf,len));
			}
	}
}

int word_init() {
	if(store_builtin_ns0("unique-word",unique_word))
		return(1);
	return(0);
}

objtype word_objtype={
	name: "word",
	tostring: word_tostring,
	dropobj: drop_word,
	evaluate: word_evaluate
};

